import math

from lightly.utils import dist
from loguru import logger


def infer_lr(base_lr: float, current_batch_size: int, base_batch_size: int, scaling: str = "linear"):
    """
    Infer learning rate set for a base_batch_size. Helps define a LR only once and batch-size can be
    varied as needed.

    base_lr: The base learning rate for a specific batch size
    base_batch_size: The batch size that `base_lr` is locked to
    current_batch_size: Batch size used in current training
    scaling: Method to scale the learning rate. linear | sqrt supported
    """
    world_size = dist.world_size()
    effective_batch_size = current_batch_size * world_size
    base_lr = base_lr
    base_lr_batch_size = base_batch_size
    scaling = scaling

    assert scaling in [
        "sqrt",
        "linear",
    ], "Only linear | sqrt scaling_types are supported"

    scale_factor = float(effective_batch_size) / base_lr_batch_size
    if scaling == "sqrt":
        scale_factor = math.pow(scale_factor, 0.5)
    scaled_lr = base_lr * scale_factor
    logger.info(f"Scaled learning rate: {scaled_lr}, {type(scaled_lr)}")
    return scaled_lr



from typing import List, Optional, Tuple, Union

import numpy as np
import SimpleITK as sitk
import torch

# https://github.com/SimpleITK/SlicerSimpleFilters/blob/master/SimpleFilters/SimpleFilters.py
SITK_INTERPOLATOR_DICT = {
    "nearest": sitk.sitkNearestNeighbor,
    "linear": sitk.sitkLinear,
    "gaussian": sitk.sitkGaussian,
    "label_gaussian": sitk.sitkLabelGaussian,
    "bspline": sitk.sitkBSpline,
    "hamming_sinc": sitk.sitkHammingWindowedSinc,
    "cosine_windowed_sinc": sitk.sitkCosineWindowedSinc,
    "welch_windowed_sinc": sitk.sitkWelchWindowedSinc,
    "lanczos_windowed_sinc": sitk.sitkLanczosWindowedSinc,
}


class SitkToTensor:
    def __init__(self, add_channel_dim: bool):
        """_summary_

        Args:
            add_channel_dim (bool):  add channel dimension to the tensor. (D)HW -> C(D)HW.
        """
        self.add_channel_dim = add_channel_dim

    def __call__(self, sitk_image):
        tensor = torch.tensor(sitk.GetArrayFromImage(sitk_image))
        return tensor.unsqueeze(0) if self.add_channel_dim else tensor


class SitkRandomSpacing:
    def __init__(
        self,
        prob: float,
        min_spacing: Union[int, List[int], Tuple[int]],
        max_spacing: Union[int, List[int], Tuple[int]],
        default_value: Union[int, float],
        tolerance: Optional[float] = None,
        interpolator: str = "linear",
    ):
        self.prob = prob
        self.min_spacing = np.array(min_spacing)
        self.max_spacing = np.array(max_spacing)
        self.default_value = default_value
        self.tolerance = None if tolerance is None else np.array(tolerance)
        self.interpolator = SITK_INTERPOLATOR_DICT[interpolator]

    def __call__(self, sitk_image):
        if torch.rand(1) > self.prob:
            return sitk_image

        current_spacing = sitk_image.GetSpacing()
        current_size = sitk_image.GetSize()

        if self.tolerance is not None:
            tolerated_min_spacing = np.array(self.min_spacing) * (1 - self.tolerance)
            tolerated_max_spacing = np.array(self.max_spacing) * (1 - self.tolerance)
        else:
            tolerated_min_spacing = self.min_spacing
            tolerated_max_spacing = self.max_spacing

        min_spacing = [max(spacings) for spacings in zip(self.min_spacing, tolerated_min_spacing)]
        max_spacing = [min(spacings) for spacings in zip(self.max_spacing, tolerated_max_spacing)]

        new_spacing = [np.random.uniform(mn, mx) for mn, mx in zip(min_spacing, max_spacing)]
        new_size = []
        for size, spacing, n_spacing in zip(current_size, current_spacing, new_spacing):
            new_size.append(int(round(size * spacing / n_spacing)))

        return sitk.Resample(
            sitk_image,
            new_size,  # size
            sitk.Transform(),  # transform
            self.interpolator,  # interpolator
            sitk_image.GetOrigin(),  # outputOrigin
            new_spacing,  # outputSpacing
            sitk_image.GetDirection(),  # outputDirection
            self.default_value,  # defaultPixelValue
            sitk_image.GetPixelID(),
        )  # outputPixelType



from typing import Any, Callable, List, Optional, Tuple

from copy import deepcopy

import torch


class Duplicate:
    """Duplicate an input and apply two different transforms. Used for SimCLR primarily."""

    def __init__(self, transforms1: Optional[Callable] = None, transforms2: Optional[Callable] = None):
        """Duplicates an input and applies the given transformations to each copy separately.

        Args:
            transforms1 (Optional[Callable], optional): _description_. Defaults to None.
            transforms2 (Optional[Callable], optional): _description_. Defaults to None.
        """
        # Wrapped into a list if it isn't one already to allow both a
        # list of transforms as well as `torchvision.transform.Compose` transforms.
        self.transforms1 = transforms1
        self.transforms2 = transforms2

    def __call__(self, input: Any) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Args:
            input (torch.Tensor or any other type supported by the given transforms): Input.

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: a tuple of two tensors.
        """
        out1, out2 = input, deepcopy(input)
        if self.transforms1 is not None:
            out1 = self.transforms1(out1)
        if self.transforms2 is not None:
            out2 = self.transforms2(out2)
        return (out1, out2)


class MultiCrop:
    """SwaV Multi-Crop augmentation."""

    def __init__(self, high_resolution_transforms: List[Callable], low_resolution_transforms: List[Callable]):
        self.high_resolution_transforms = high_resolution_transforms
        self.low_resolution_transforms = low_resolution_transforms

    def __call__(self, input):
        high_resolution_crops = [transform(input) for transform in self.high_resolution_transforms]
        low_resolution_crops = [transform(input) for transform in self.low_resolution_transforms]
        return (high_resolution_crops, low_resolution_crops)



from typing import Any, Dict

import sys
from functools import partial

import fire
import yaml
from loguru import logger
from monai.bundle.scripts import run
from monai.utils.misc import ensure_tuple

from lighter.utils.dynamic_imports import import_module_from_path

# Trainer methods calls in YAML format with support for command line arguments.
# Waiting for https://github.com/Project-MONAI/MONAI/pull/5854#issuecomment-1384800886.
trainer_methods = {
    "fit": yaml.safe_load(
        """
        fit:
            _method_: >
                $@trainer.fit(model=@fit#model,
                              ckpt_path=@fit#ckpt_path)
            model: "@system"
            ckpt_path: null"""
    ),
    "validate": yaml.safe_load(
        """
        validate:
            _method_: > 
                $@trainer.validate(model=@validate#model,
                                   ckpt_path=@validate#ckpt_path,
                                   verbose=@validate#verbose)
            model: "@system"
            ckpt_path: null
            verbose: True"""
    ),
    "predict": yaml.safe_load(
        """
        predict:
            _method_: >
                $@trainer.predict(model=@predict#model,
                                  ckpt_path=@predict#ckpt_path)
            model: "@system"
            ckpt_path: null"""
    ),
    "test": yaml.safe_load(
        """
        test:
            _method_: >
                $@trainer.test(model=@test#model,
                               ckpt_path=@test#ckpt_path,
                               verbose=@test#verbose)
            model: "@system"
            ckpt_path: null
            verbose: True"""
    ),
    "tune": yaml.safe_load(
        """
        tune:
            _method_: > 
                $@trainer.tune(model=@tune#model,
                               ckpt_path=@tune#ckpt_path,
                               scale_batch_size_kwargs=@tune#scale_batch_size_kwargs,
                               lr_find_kwargs=@tune#lr_find_kwargs,
                               method=@tune#method)
            model: "@system"
            ckpt_path: null
            scale_batch_size_kwargs: null
            lr_find_kwargs: null
            method: fit
    """
    ),
}


def interface():
    fire.Fire({k: partial(run_trainer_method, v) for k, v in trainer_methods.items()})


def run_trainer_method(trainer_method: Dict, **kwargs: Any):
    """Call monai.bundle.run() on a Trainer method. If a project path
    is defined in the config file(s), import it.

    Args:
        trainer_method (Dict): Trainer method to run, collection defined in `trainer_methods`
        **kwargs (Any): keyword arguments for the monai.bundle.run function.
    """
    if "config_file" in kwargs:
        project_imported = False
        # Handle multiple configs
        for config in ensure_tuple(kwargs["config_file"]):
            config = yaml.safe_load(open(config))
            # Import the project as a module
            if "project" in config:
                # Only one config file can specify the project path
                if project_imported:
                    logger.error("`project` must be specified in one config only. Exiting.")
                    sys.exit()
                import_module_from_path("project", config["project"])
                project_imported = True
    # Run the Trainer method
    run(**trainer_method, **kwargs)



import torch
from lightly.loss.memory_bank import MemoryBankModule
from lightly.models.modules import SwaVProjectionHead, SwaVPrototypes
from torch import nn


class SwaV(nn.Module):
    def __init__(
        self,
        backbone,
        num_ftrs,
        out_dim,
        n_prototypes,
        n_queues,
        queue_length=0,
        start_queue_at_epoch=0,
        n_steps_frozen_prototypes=0,
    ):
        super().__init__()
        self.backbone = backbone
        self.projection_head = SwaVProjectionHead(num_ftrs, num_ftrs, out_dim)
        self.prototypes = SwaVPrototypes(out_dim, n_prototypes, n_steps_frozen_prototypes)

        self.queues = None
        if n_queues > 0:
            self.queues = [MemoryBankModule(size=queue_length) for _ in range(n_queues)]
            self.queues = nn.ModuleList(self.queues)
            self.queue_length = queue_length
            self.num_features_queued = 0
            self.start_queue_at_epoch = start_queue_at_epoch

    def forward(self, high_resolution, low_resolution, epoch=None, step=None):
        self.prototypes.normalize()

        high_resolution_features = [self._subforward(x) for x in high_resolution]
        low_resolution_features = [self._subforward(x) for x in low_resolution]

        high_resolution_prototypes = [self.prototypes(x, step) for x in high_resolution_features]
        low_resolution_prototypes = [self.prototypes(x, step) for x in low_resolution_features]
        queue_prototypes = self._get_queue_prototypes(high_resolution_features, epoch)

        return high_resolution_prototypes, low_resolution_prototypes, queue_prototypes

    def _subforward(self, input):
        features = self.backbone(input).flatten(start_dim=1)
        features = self.projection_head(features)
        features = nn.functional.normalize(features, dim=1, p=2)
        return features

    @torch.no_grad()
    def _get_queue_prototypes(self, high_resolution_features, epoch=None):
        if self.queues is None:
            return None

        if len(high_resolution_features) != len(self.queues):
            raise ValueError(
                f"The number of queues ({len(self.queues)}) should be equal to the number of high "
                f"resolution inputs ({len(high_resolution_features)}). Set `n_queues` accordingly."
            )

        # Get the queue features
        queue_features = []
        for i in range(len(self.queues)):
            _, features = self.queues[i](high_resolution_features[i], update=True)
            # Queue features are in (num_ftrs X queue_length) shape, while the high res
            # features are in (batch_size X num_ftrs). Swap the axes for interoperability.
            features = torch.permute(features, (1, 0))
            queue_features.append(features)

        # Do not return queue prototypes if not enough features have been queued
        self.num_features_queued += high_resolution_features[0].shape[0]
        if self.num_features_queued < self.queue_length:
            return None

        # If loss calculation with queue prototypes starts at a later epoch,
        # just queue the features and return None instead of queue prototypes.
        if self.start_queue_at_epoch > 0:
            if epoch is None:
                raise ValueError(
                    "The epoch number must be passed to the `forward()` " "method if `start_queue_at_epoch` is greater than 0."
                )
            if epoch < self.start_queue_at_epoch:
                return None

        # Assign prototypes
        queue_prototypes = [self.prototypes(x) for x in queue_features]
        return queue_prototypes



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lighter',
 'lighter.callbacks',
 'lighter.contrib',
 'lighter.contrib.models',
 'lighter.contrib.optimizer',
 'lighter.contrib.transforms',
 'lighter.contrib.utils',
 'lighter.utils']

package_data = \
{'': ['*']}

install_requires = \
['fire>=0.5.0,<0.6.0',
 'lightly>=1.2.43,<2.0.0',
 'loguru>=0.6.0,<0.7.0',
 'monai-weekly>=1.2.dev2304,<2.0',
 'pandas>=1.5.3,<2.0.0',
 'py>=1.11.0,<2.0.0',
 'pytorch-lightning>=1.9.0,<2.0.0',
 'tensorboard>=2.11.2,<3.0.0',
 'torch>=1.13.1,<2.0.0',
 'torchmetrics>=0.11.0,<0.12.0',
 'torchvision>=0.14.1,<0.15.0']

entry_points = \
{'console_scripts': ['lighter = lighter.utils.cli:interface']}

setup_kwargs = {
    'name': 'project-lighter',
    'version': '0.0.1',
    'description': 'YAML-based automated rapid prototyping framework for deep learning experiments',
    'long_description': '# Lighter\n[![build](https://github.com/project-lighter/lighter/actions/workflows/build.yml/badge.svg?branch=main)](https://github.com/project-lighter/lighter/actions/workflows/build.yml) ![Coverage](./assets/images/coverage.svg)\n\n\n## Setup\n\nRun the following commands to develop with the repository and add the package in develop mode to your environment\n\n````\nmake setup\nmake install \n````\n\nTo setup the pre-commit hook to format code before commit,\n````\nmake pre-commit-install\n````\n',
    'author': 'Ibrahim Hadzic',
    'author_email': 'ibrahimhadzic45@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/lighter/lighter',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import Any, Dict, List, Optional, Tuple, Union

import re
import sys
from datetime import datetime
from pathlib import Path

import torch
import torch.distributed as dist
import torchvision
from loguru import logger
from monai.utils.module import optional_import
from pytorch_lightning import Callback, Trainer
from torch.utils import tensorboard
from yaml import safe_load

from lighter import LighterSystem

LIGHTNING_TO_LIGHTER_STAGE = {"train": "train", "validate": "val", "test": "test"}
OPTIONAL_IMPORTS = {}


class LighterLogger(Callback):
    def __init__(
        self,
        project,
        log_dir,
        tensorboard=False,
        wandb=False,
        input_type=None,
        target_type=None,
        pred_type=None,
        max_samples=None,
    ) -> None:
        self.project = project
        # Only used on rank 0, the dir is created in setup().
        self.log_dir = Path(log_dir) / project / datetime.now().strftime("%Y%m%d_%H%M%S")

        self.data_types = {"input": input_type, "target": target_type, "pred": pred_type}
        # Max number of samples from the batch to log.
        self.max_samples = max_samples

        self.tensorboard = tensorboard
        self.wandb = wandb

        # Running loss. Resets at each epoch. Loss is not calculated for `test` mode.
        self.loss = {"train": 0, "val": 0}
        # Epoch steps. Resets at each epoch. No `test` mode step counter as loss is not calculated for it.
        self.epoch_step_counter = {"train": 0, "val": 0}
        # Global steps. Replaces `Trainer.global_step` because it counts optimizer steps
        # instead of batch steps, which can be problematic when using gradient accumulation.
        self.global_step_counter = {"train": 0, "val": 0, "test": 0}

    def setup(self, trainer: Trainer, pl_module: LighterSystem, stage: str) -> None:
        """
        Initialize logging for the LighterSystem.
        TODO: improve this docstring.

        Args:
            trainer (Trainer): Trainer, passed automatically by PyTorch Lightning.
            pl_module (LighterSystem): LighterSystem, passed automatically by PyTorch Lightning.
            stage (str): stage of the training process. Passed automatically by PyTorch Lightning.
        """
        if trainer.logger is not None:
            logger.error("When using LighterLogger, set Trainer(logger=None).")
            sys.exit()

        if dist.is_initialized() and dist.get_rank() != 0:
            return

        self.log_dir.mkdir(parents=True)

        # Load the dumped config file to log it to the loggers.
        # config = safe_load(open(self.log_dir / "config.yaml"))

        # Loguru log file.
        # logger.add(sink=self.log_dir / f"{stage}.log")

        # Tensorboard initialization.
        if self.tensorboard:
            tensorboard_dir = self.log_dir / "tensorboard"
            tensorboard_dir.mkdir()
            self.tensorboard = tensorboard.SummaryWriter(log_dir=tensorboard_dir)
            # self.tensorboard.add_hparams(config)

        # Wandb initialization.
        if self.wandb:
            OPTIONAL_IMPORTS["wandb"], wandb_available = optional_import("wandb")
            if not wandb_available:
                logger.error("Weights & Biases not installed. To install it, run `pip install wandb`. Exiting.")
                sys.exit()
            wandb_dir = self.log_dir / "wandb"
            wandb_dir.mkdir()
            self.wandb = OPTIONAL_IMPORTS["wandb"].init(project=self.project, dir=wandb_dir)
            # self.wandb.config.update(config)

    def teardown(self, trainer: Trainer, pl_module: LighterSystem, stage: str) -> None:
        if dist.is_initialized() and dist.get_rank() != 0:
            return
        self.tensorboard.close()

    def _log(self, outputs: dict, mode: str, global_step: int, is_epoch=False) -> None:
        """Logs the outputs to TensorBoard and Weights & Biases (if enabled).
        The outputs are logged as scalars and images, depending on the configuration.

        Args:
            outputs (dict): model outputs.
            mode (str): current mode (train/val/test).
            global_step (int): current global step.
            is_epoch (bool): whether the log is being done at the end
                of an epoch or astep. Default is False.
        """
        if dist.is_initialized() and dist.get_rank() != 0:
            return

        step_or_epoch = "epoch" if is_epoch else "step"

        # Loss
        if outputs["loss"] is not None:
            name = f"{mode}/loss/{step_or_epoch}"
            self._log_scalar(name, outputs["loss"], global_step)

        # Metrics
        if outputs["metrics"] is not None:
            for name, metric in outputs["metrics"].items():
                name = f"{mode}/metrics/{name}_{step_or_epoch}"
                self._log_scalar(name, metric, global_step)

        # Epoch does not log input, target, and pred.
        if is_epoch:
            return

        # Input, Target, Pred
        for data_name in ["input", "target", "pred"]:
            if self.data_types[data_name] is None:
                continue

            data_type = self.data_types[data_name]
            data = outputs[data_name]
            name = f"{mode}/data/{data_name}_{step_or_epoch}"

            # Scalar
            if data_type == "scalar":
                self._log_scalar(name, data, global_step)

            # Image
            elif data_type == "image":
                # Check if the data type is valid.
                check_image_data_type(data, data_name)
                for identifier, image in parse_image_data(data):
                    name = name if identifier is None else f"{name}_{identifier}"
                    # Slice to `max_samples` only if it less than the batch size.
                    if self.max_samples is not None and self.max_samples < image.shape[0]:
                        image = image[: self.max_samples]
                    # Preprocess a batch of images into a single, loggable, image.
                    image = preprocess_image(image)
                    self._log_image(name, image, global_step)
            else:
                logger.error(f"`{data_name}_type` does not support `{data_type}`.")
                sys.exit()

    def _log_scalar(self, name: str, scalar: Union[int, float, torch.Tensor], global_step: int) -> None:
        """Logs the scalar to TensorBoard and Weights & Biases (if enabled).

        Args:
            name (str): name of the image to be logged.
            scalar (Union[int, float, torch.Tensor]): image to be logged.
            global_step (int): current global step.
        """
        if not isinstance(scalar, (int, float, torch.Tensor)):
            raise NotImplementedError("LighterLogger currently supports only single scalars.")
        if isinstance(scalar, torch.Tensor) and scalar.dim() > 0:
            raise NotImplementedError("LighterLogger currently supports only single scalars.")

        if self.tensorboard:
            self.tensorboard.add_scalar(name, scalar, global_step=global_step)
        if self.wandb:
            self.wandb.log({name: scalar}, step=global_step)

    def _log_image(self, name: str, image: torch.Tensor, global_step: int) -> None:
        """Logs the image to TensorBoard and Weights & Biases (if enabled).

        Args:
            name (str): name of the image to be logged.
            image (torch.Tensor): image to be logged.
            global_step (int): current global step.
        """
        if self.tensorboard:
            self.tensorboard.add_image(name, image, global_step=global_step)
        if self.wandb:
            self.wandb.log({name: OPTIONAL_IMPORTS["wandb"].Image(image)}, step=global_step)

    def _on_batch_end(self, outputs: Dict, trainer: Trainer) -> None:
        """Performs logging at the end of a batch/step. It logs the loss and metrics,
        and, if specified how, the input, target, and pred data.

        Args:
            outputs (Dict): output dict from the model.
            trainer (Trainer): Trainer, passed automatically by PyTorch Lightning.
        """
        if not trainer.sanity_checking:
            mode = LIGHTNING_TO_LIGHTER_STAGE[trainer.state.stage]
            # Accumulate the loss.
            if mode in ["train", "val"]:
                self.loss[mode] += outputs["loss"].item()
            # Logging frequency.
            if self.global_step_counter[mode] % trainer.log_every_n_steps == 0:
                # Log. Done only on rank 0.
                self._log(outputs, mode, global_step=self._get_global_step(trainer))
            # Increment the step counters.
            self.global_step_counter[mode] += 1
            if mode in ["train", "val"]:
                self.epoch_step_counter[mode] += 1

    def _on_epoch_end(self, trainer: Trainer, pl_module: LighterSystem) -> None:
        """Performs logging at the end of an epoch. It calculates the average
        loss and metrics for the epoch and logs them. In distributed mode, it averages
        the losses and metrics from all processes.

        Args:
            trainer (Trainer): Trainer, passed automatically by PyTorch Lightning.
            pl_module (LighterSystem): LighterSystem, passed automatically by PyTorch Lightning.
        """
        if not trainer.sanity_checking:
            mode = LIGHTNING_TO_LIGHTER_STAGE[trainer.state.stage]
            outputs = {"loss": None, "metrics": None}

            # Loss
            if mode in ["train", "val"]:
                # Get the accumulated loss.
                loss = self.loss[mode]
                # Reduce the loss to rank 0 and average it.
                if dist.is_initialized():
                    # Distributed communication works only tensors.
                    loss = torch.tensor(loss).to(pl_module.device)
                    # On rank 0, sum the losses from all ranks. Other ranks remain with the same loss as before.
                    dist.reduce(loss, dst=0)
                    # On rank 0, average the loss sum by dividing it with the number of processes.
                    loss = loss.item() / dist.get_world_size() if dist.get_rank() == 0 else loss.item()
                # Divide the accumulated loss by the number of steps in the epoch.
                loss /= self.epoch_step_counter[mode]
                outputs["loss"] = loss

            # Metrics
            # Get the torchmetrics.
            metrics = getattr(pl_module, f"{mode}_metrics")
            if metrics is not None:
                # Compute the epoch metrics.
                outputs["metrics"] = metrics.compute()
                # Reset the metrics for the next epoch.
                metrics.reset()

            # Log. Done only on rank 0.
            self._log(outputs, mode, is_epoch=True, global_step=self._get_global_step(trainer))

    def _get_global_step(self, trainer: Trainer) -> int:
        """Return the global step for the current mode. Note that when Trainer
        is running Trainer.fit() and is in `val` mode, this method will return
        the global step of the `train` mode in order to correctly log the validation
        steps against training steps.

        Args:
            trainer (Trainer): Trainer, passed automatically by PyTorch Lightning.

        Returns:
            int: global step.
        """
        mode = LIGHTNING_TO_LIGHTER_STAGE[trainer.state.stage]
        # When validating in Trainer.fit(), return the train steps instead of the
        # val steps to correctly
        if mode == "val" and trainer.state.fn == "fit":
            return self.global_step_counter["train"]
        return self.global_step_counter[mode]

    def on_train_epoch_start(self, trainer: Trainer, pl_module: LighterSystem) -> None:
        # Reset the loss and the epoch step counter for the next epoch.
        self.loss["train"] = 0
        self.epoch_step_counter["train"] = 0

    def on_validation_epoch_start(self, trainer: Trainer, pl_module: LighterSystem) -> None:
        # Reset the loss and the epoch step counter for the next epoch.
        self.loss["val"] = 0
        self.epoch_step_counter["val"] = 0

    def on_train_batch_end(self, trainer: Trainer, pl_module: LighterSystem, outputs: Any, batch: Any, batch_idx: int) -> None:
        self._on_batch_end(outputs, trainer)

    def on_validation_batch_end(
        self, trainer: Trainer, pl_module: LighterSystem, outputs: Any, batch: Any, batch_idx: int, dataloader_idx: int
    ) -> None:
        self._on_batch_end(outputs, trainer)

    def on_test_batch_end(
        self, trainer: Trainer, pl_module: LighterSystem, outputs: Any, batch: Any, batch_idx: int, dataloader_idx: int
    ) -> None:
        self._on_batch_end(outputs, trainer)

    def on_train_epoch_end(self, trainer: Trainer, pl_module: LighterSystem) -> None:
        self._on_epoch_end(trainer, pl_module)

    def on_validation_epoch_end(self, trainer: Trainer, pl_module: LighterSystem) -> None:
        self._on_epoch_end(trainer, pl_module)

    def on_test_epoch_end(self, trainer: Trainer, pl_module: LighterSystem) -> None:
        self._on_epoch_end(trainer, pl_module)


def preprocess_image(image: torch.Tensor) -> torch.Tensor:
    """Preprocess the image before logging it. If it is a batch of multiple images,
    it will create a grid image of them. In case of 3D, a single image is displayed
    with slices stacked vertically, while a batch as a grid where each column is
    a different 3D image.
    Args:
        image (torch.Tensor): 2D or 3D image tensor.
    Returns:
        torch.Tensor: image ready for logging.
    """
    image = image.detach().cpu()
    # 3D image (NCDHW)
    has_three_dims = image.ndim == 5
    if has_three_dims:
        # Reshape 3D image from NCDHW to NC(D*H)W format
        shape = image.shape
        image = image.view(shape[0], shape[1], shape[2] * shape[3], shape[4])
    if image.shape[0] == 1:
        image = image[0]
    else:
        # If more than one image, create a grid
        nrow = image.shape[0] if has_three_dims else 8
        image = torchvision.utils.make_grid(image, nrow=nrow)
    return image


def check_image_data_type(data: Any, name: str) -> None:
    """Check the input image data for its type. Valid image data types are:
        - torch.Tensor
        - List[torch.Tensor]
        - Dict[str, torch.Tensor]
        - Dict[str, List[torch.Tensor]]

    Args:
        data (Any): image data to check
        name (str): name of the image data, for logging purposes.
    """
    if isinstance(data, dict):
        is_valid = all(check_image_data_type(elem) for elem in data.values())
    elif isinstance(data, list):
        is_valid = all(check_image_data_type(elem) for elem in data)
    elif isinstance(data, torch.Tensor):
        is_valid = True
    else:
        is_valid = False

    if not is_valid:
        logger.error(
            f"`{name}` has to be a Tensor, List[Tensors], Dict[str, Tensor]"
            f", or Dict[str, List[Tensor]]. `{type(data)}` is not supported."
        )
        sys.exit()


def parse_image_data(
    data: Union[Dict[str, torch.Tensor], Dict[str, List[torch.Tensor]], List[torch.Tensor], torch.Tensor]
) -> List[Tuple[Optional[str], torch.Tensor]]:
    """Given input data, this function will parse it and return a list of tuples where
    each tuple contains an identifier and a tensor.

    Args:
        data (Union[Dict[str, torch.Tensor], Dict[str, List[torch.Tensor]], List[torch.Tensor], torch.Tensor]): image tensor(s).

    Returns:
        List[Tuple[Optional[str], torch.Tensor]]: a list of tuples where the first element is
            a string identifier (or `None` if there is only one image) and the second an image tensor.
    """
    result = []
    if isinstance(data, dict):
        for key, value in data.items():
            if isinstance(value, list):
                for i, tensor in enumerate(value):
                    result.append((f"{key}_{i}", tensor) if len(value > 1) else (key, tensor))
            else:
                result.append((key, value))
    elif isinstance(data, list):
        for i, tensor in enumerate(data):
            result.append((str(i), tensor))
    else:
        result.append((None, data))
    return result



import torch
from torch.nn import Identity, Module, Sequential

from lighter.utils.misc import dot_notation_setattr


def reshape_pred_if_single_value_prediction(pred: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
    """When the task is to predict a single value, pred and target dimensions often
    mismatch - dataloader returns the value in the (B) shape, while the network
    returns predictions in the (B, 1) shape, where the second dim is redundant.

    Args:
        pred (torch.Tensor): predicted tensor.
        target (torch.Tensor): target tensor.

    Returns:
        torch.Tensor: reshaped predicted tensor if that was necessary.
    """
    if isinstance(pred, torch.Tensor) and target is not None:
        if len(pred.shape) == 2 and len(target.shape) == 1 == pred.shape[1]:
            return pred.flatten()
    return pred


def replace_layer_with(model: Module, layer_name: str, new_layer: Module) -> Module:
    """Replaces the specified layer of the network with another layer.

    Args:
        model (Module): PyTorch model to be edited
        layer_name (string): Name of the layer which will be replaced.
            Dot-notation supported, e.g. "layer10.fc".

    Returns:
        Module: PyTorch model with the new layer set at the specified location.
    """
    dot_notation_setattr(model, layer_name, new_layer)
    return model


def replace_layer_with_identity(model: Module, layer_name: str) -> Module:
    """Replaces any layer of the network with an Identity layer.
    Useful for removing the last layer of a network to be used as a backbone
    of an SSL model.

    Args:
        model (Module): PyTorch model to be edited
        layer_name (string): Name of the layer which will be replaced with an
            Identity function. Dot-notation supported, e.g. "layer10.fc".

    Returns:
        Module: PyTorch model with Identity layer at the specified location.
    """
    return replace_layer_with(model, layer_name, Identity())


def remove_last_layer_sequentially(model: Module()) -> Sequential:
    """Removes the last layer of a network and returns it as an nn.Sequential model.
    Useful when a network is to be used as a backbone of an SSL model.

    Args:
        model (Module): PyTorch model object.

    Returns:
        Sequential: PyTorch Sequential model with the last layer removed.
    """
    return Sequential(*list(model.children())[:-1])



import random

import torch
from torch.utils.data import DataLoader


def collate_fn_replace_corrupted(batch: torch.Tensor, dataset: DataLoader) -> torch.Tensor:
    """Collate function that allows to replace corrupted examples in the batch.
    The dataloader should return `None` when that occurs.
    The `None`s in the batch are replaced with other, randomly-selected, examples.

    Args:
        batch (torch.Tensor): batch from the DataLoader.
        dataset (Dataset): dataset that the DataLoader is passing through. Needs to be fixed
            in place with functools.partial before passing it to DataLoader's 'collate_fn' option
            as 'collate_fn' should only have a single argument - batch. Example:
                ```
                collate_fn = functools.partial(collate_fn_replace_corrupted, dataset=dataset)`
                loader = DataLoader(dataset, ..., collate_fn=collate_fn).
                ```
    Returns:
        torch.Tensor: batch with new examples instead of corrupted ones.
    """
    # Idea from https://stackoverflow.com/a/57882783
    original_batch_len = len(batch)
    # Filter out all the Nones (corrupted examples)
    batch = list(filter(lambda x: x is not None, batch))
    filtered_batch_len = len(batch)
    # Num of corrupted examples
    num_corrupted = original_batch_len - filtered_batch_len
    if num_corrupted > 0:
        # Replace a corrupted example with another randomly selected example
        batch.extend([dataset[random.randint(0, len(dataset))] for _ in range(num_corrupted)])
        # Recursive call to replace the replacements if they are corrupted
        return collate_fn_replace_corrupted(batch, dataset)
    # Finally, when the whole batch is fine, return it
    return torch.utils.data.dataloader.default_collate(batch)
